#!/usr/bin/env python

import gzip
import codecs
import random
import csv
from io import reader, writer

if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-o", "--output", dest="output")
    parser.add_argument(nargs="+", dest="inputs")
    options = parser.parse_args()

    key_fields = ["task", "fold", "model", "size"]
    fields = set()
    rows = {}
    for ifile in options.inputs:
        with gzip.open(ifile) as ifd:
            for entry in csv.DictReader(ifd, delimiter="\t"):
                key = tuple([entry[f] for f in key_fields])
                rows[key] = rows.get(key, {})
                for k, v in entry.iteritems():
                    fields.add(k)
                    if k not in key_fields:
                        rows[key][k] = v
                    

    with gzip.open(options.output, "w") as ofd:
        c = csv.DictWriter(ofd, fieldnames=fields, delimiter="\t")
        c.writeheader()
        for fs, r in rows.iteritems():
            c.writerow({k : v for k, v in zip(key_fields, fs) + list(r.iteritems())})




if __name__ == "__main__":

    import flask
    import argparse
    from glob import glob
    import os
    import os.path
    import logging
    import subprocess
    from pkg_resources import resource_string, resource_listdir, resource_isdir
    import json
    import re
    import gzip

    def get_files(path):
        entries = resource_listdir("steamroller", path)
        files = [os.path.join(path, e) for e in entries if
                 not resource_isdir("steamroller", e)
                 and e.endswith("py")
                 and os.path.basename(e) != "resources.py"]
        dirs = [e for e in entries if resource_isdir("steamroller", e)]
        return files + sum([get_files(os.path.join(path, d)) for d in dirs], [])

    def to_texts(filenames, num):
        text = "\n".join([resource_string("steamroller", f) for f in filenames])
        lines = [l for l in text.split("\n") if not re.match(r"^\s*$", l)]
        per = len(lines) / num
        return [re.sub(r"\s", " ", "\n".join(lines[i * per:(i + 1) * per])) for i in range(num)]

    def init(args, _):
        sconstruct = resource_string(__name__, "data/SConstruct")
        steamroller_config = resource_string(__name__, "data/steamroller_config.json.template")
        python_docs = [re.sub(r"\s", " ", str(getattr(__builtins__, a).__doc__))
                       for a in dir(__builtins__)]
        python_code = to_texts(get_files("/"), len(python_docs))
        try:
            os.mkdir("tasks")
        except OSError:
            logging.info("tasks/ directory already exists")
        try:
            os.mkdir("csv")
        except OSError:
            logging.info("csv/ directory already exists")
        with gzip.open("csv/example.txt.gz", "w") as ofd:
            for i, t in enumerate(python_docs):
                ofd.write("%d\tdoc\t%s\n" % (i, t))
            for i, t in enumerate(python_code):
                ofd.write("%d\tcode\t%s\n" % (len(python_docs) + i, t))
        subprocess.call(["python",
                         "-m",
                         "steamroller.tools.convert",
                         "-i",
                         "csv/example.txt.gz",
                         "-t",
                         "attribute",
                         "-o",
                         "tasks/example.tgz"])
        if (not args.force) and (os.path.exists("SConstruct") or os.path.exists("steamroller_config.py")):
            logging.error("Refusing to overwrite existing SConstruct or steamroller_config.json files (try \"--force\")")
        else:
            with open("SConstruct", "w") as ofd:
                ofd.write(sconstruct)
            with open("steamroller_config.json", "w") as ofd:
                ofd.write(steamroller_config)

    def run(args, scons_args):
        subprocess.call(["scons"] + scons_args + ["CONFIG_FILE=%s" % args.config])

    def serve(args, _):
        app = flask.Flask("SteamRoller")
        with open(args.config) as ifd:
            config = json.load(ifd)
        task_names = [x["NAME"] for x in config["TASKS"]]
        @app.route("/")
        def browse():
            return "<html><body><h1>{}</h1><ul>{}</ul></body></html>".format("SteamRoller results browser",
                                                                             "\n".join(["<li><a href=\"/tasks/{}\">{}</a></li>".format(t, t) for t in task_names]))

        @app.route("/tasks/<task>")
        def experiment(task):
            images = glob("work/{}*png".format(task))
            return "<html><body><h3><a href=\"/\">Back</a></h3><h1>{}</h1>{}</body></html>".format(task, "\n".join(["<img src=\"/{}\"/>".format(i) for i in images]))

        @app.route("/work/<image_file>")
        def image(image_file):
            with open(os.path.join("work", image_file)) as ifd:
                return ifd.read()

        app.run(port=options.port, host=options.host)


    parser = argparse.ArgumentParser("steamroller")
    subparsers = parser.add_subparsers(help="sub-commands")
    init_parser = subparsers.add_parser("init", help="Initialize an experiment directory")
    init_parser.add_argument("-f", "--force", dest="force", default=False, action="store_true", help="Overwrite existing files")
    init_parser.set_defaults(func=init)
    run_parser = subparsers.add_parser("run", help="Run experiments by calling SCons (additional arguments are passed through", add_help=False)
    run_parser.add_argument("-c", "--config", dest="config", default="steamroller_config.json")
    run_parser.set_defaults(func=run)
    serve_parser = subparsers.add_parser("serve", help="Serve experiment results")
    serve_parser.add_argument("-p", "--port", dest="port", default=8080, type=int)
    serve_parser.add_argument("-H", "--host", dest="host", default="localhost")
    serve_parser.add_argument("--config", dest="config", default="steamroller_config.json")
    serve_parser.set_defaults(func=serve)
    options, rest = parser.parse_known_args()

    logging.basicConfig(level=logging.INFO)

    options.func(options, rest)



import functools
from SCons.Action import Action
from SCons.Builder import Builder
import SCons.Util
import subprocess
import logging
import time
import shlex

def qsub(command, name, std, dep_ids=[], grid_resources=[]):
    deps = "" if len(dep_ids) == 0 else "-hold_jid {}".format(",".join([str(x) for x in dep_ids]))
    res = "" if len(grid_resources) == 0 else "-l {}".format(",".join([str(x) for x in grid_resources]))
    qcommand = "qsub -shell n -V -N {} -b y -cwd -j y -terse -o {} {} {} ".format(name, std, deps, res) + command
    logging.info(qcommand)
    p = subprocess.Popen(shlex.split(qcommand), stdout=subprocess.PIPE)
    out, err = p.communicate()
    return int(out.strip())


def make_emitter(script):
    def emitter(target, source, env):
        return (target + [target[0].rstr() + ".resources.txt"], source)
    return emitter


def make_builder(env, base_command, name):
    timed_command = "/usr/bin/time --verbose -o ${TARGETS[-1]} " + base_command
    def grid_method(target, source, env):
        depends_on = set(filter(lambda x : x != None, [s.GetTag("built_by_job") for s in source]))
        job_id = qsub(env.subst(timed_command, source=source, target=target), name, "{}.qout".format(target[-1].rstr()), depends_on, env["GRID_RESOURCES"])
        for t in target:
            t.Tag("built_by_job", job_id)
        logging.info("Job %d depends on %s", job_id, depends_on)
        return None

    return Builder(action=Action(grid_method, "grid(" + base_command + ")") if env["GRID"] else Action(timed_command, base_command),
                   emitter=make_emitter(base_command.split()[0]))


def generate(env):

    env.SetDefault(
        MAX_NGRAM=4,
    )

    for name, command in [
            ("GetCount", "python -m steamroller.tools.count --input ${SOURCES[0]} --output ${TARGETS[0]}"),
            ("CreateSplit", "python -m steamroller.tools.split --total_file ${SOURCES[0]} --training_size ${TRAINING_SIZE} --testing_size ${TESTING_SIZE} --train ${TARGETS[0]} --test ${TARGETS[1]}"),
            ("Accuracy", "python -m steamroller.metrics.accuracy -o ${TARGETS[0]} ${SOURCES}"),
            ("FScore", "python -m steamroller.metrics.fscore -o ${TARGETS[0]} ${SOURCES}"),            
            ("CollateResources", "python -m steamroller.tools.resources -o ${TARGETS[0]} -s ${STAGE} ${SOURCES}"),
            ("CombineCSVs", "python -m steamroller.tools.combine_csvs -o ${TARGETS[0]} ${SOURCES}"),
            ("ModelSizes", "python -m steamroller.tools.model_sizes -o ${TARGETS[0]} ${SOURCES}"),        
            ("Plot", "python -m steamroller.plots.${TYPE} --output ${TARGETS[0]} --x ${X} --y ${Y} --xlabel \"${XLABEL}\" --ylabel \"${YLABEL}\" --title \"'${TITLE}'\" --input ${SOURCES[0]} --color \"${COLOR}\" --color_label \"'${COLOR_LABEL}'\""),
    ]:
        env["BUILDERS"][name] = make_builder(env, command, name)
        
    for model in env["MODELS"]:
        if not model.get("DISABLED", False):
            env["BUILDERS"]["Train{}".format(model["NAME"])] = make_builder(env, model["TRAIN_COMMAND"], "Train{}".format(model["NAME"]))
            env["BUILDERS"]["Apply{}".format(model["NAME"])] = make_builder(env, model["APPLY_COMMAND"], "Apply{}".format(model["NAME"]))

    def wait_for_grid(target, source, env):
        while True:
            p = subprocess.Popen(["qstat"], stdout=subprocess.PIPE)
            out, err = p.communicate()   
            lines = out.strip().split("\n")
            if len(lines) < 2:
                break
            else:
                counts = {}
                for line in [l for l in lines if l.startswith(" ")]:
                    toks = line.strip().split()
                    counts[toks[4][0]] = counts.get(toks[4][0], 0) + 1
            logging.info("Running: %d Waiting: %d Held: %d", counts.get("r", 0), counts.get("w", 0), counts.get("h", 0))
            time.sleep(env["GRID_CHECK_INTERVAL"])
        return None

    env["BUILDERS"]["WaitForGrid"] = Builder(action=wait_for_grid)


def exists(env):
    return 1



from concrete import Communication, AnnotationMetadata, CommunicationTagging, Section, Sentence, TextSpan, CommunicationTagging
import gzip
import codecs
from concrete.util.file_io import CommunicationWriterTGZ
from concrete.util.concrete_uuid import AnalyticUUIDGeneratorFactory
from time import time

writer = codecs.getwriter("utf-8")
reader = codecs.getreader("utf-8")

if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", dest="input", help="Input file where each line is \"ID<tab>TAG<tab>TEXT\"")
    parser.add_argument("-o", "--output", dest="output", help="Tar file to write Communications to")
    parser.add_argument("-t", "--tag_type", dest="tag_type", default=None, help="Type of tag (e.g. \"language\"): defaults to None, in which case the tag column is ignored (but must still be present!)")
    options = parser.parse_args()

    ugf = AnalyticUUIDGeneratorFactory()
    ofd = CommunicationWriterTGZ(options.output)
    with reader(gzip.open(options.input)) as ifd:
        for i, line in enumerate(ifd):
            cid, label, text = line.strip().split("\t")
            g = ugf.create()
            t = int(time())
            comm = Communication(id=cid,
                                 uuid=g.next(),
                                 type="Text document",
                                 text=text,
                                 communicationTaggingList=[CommunicationTagging(uuid=g.next(),
                                                                                metadata=AnnotationMetadata(tool="Gold labeling",
                                                                                                            timestamp=t,
                                                                                                            kBest=1,
                                                                                ),
                                                                                taggingType=options.tag_type,
                                                                                tagList=[label],
                                                                                confidenceList=[1.0],
                                 )],
                                 metadata=AnnotationMetadata(tool="text_to_concrete.py ingester", timestamp=t, kBest=1),
                                 sectionList=[Section(uuid=g.next(),
                                                      textSpan=TextSpan(start=0, ending=len(text)),
                                                      kind="content",
                                                      )
                                 ])
            ofd.write(comm)
    ofd.close()



from concrete.util.file_io import CommunicationReader
import gzip
import codecs
import os.path
import logging


logging.basicConfig(level=logging.INFO)


writer = codecs.getwriter("utf-8")
reader = codecs.getreader("utf-8")


def read_data(data_file, num_file, tag_type="attribute"):
    conc = data_file.endswith("tgz")
    nums = set()
    with gzip.open(num_file) as ifd:
        for n in ifd:
            nums.add(int(n.rstrip("\n")))
    items = []
    ifd = CommunicationReader(data_file) if conc else reader(gzip.open(data_file))
    for n, item in enumerate(ifd):
        if n in nums:
            if conc:
                text = item[0].text
                cid = item[0].id
                label = [t for t in item[0].communicationTaggingList if
                         t.taggingType == tag_type and t.metadata.tool == "Gold labeling"][0].tagList[0]
            else:
                cid, label, text = item.rstrip("\n").split("\t")
            items.append((cid, label, text))
    logging.info("Read %d Communications", len(items))
    return items

def write_probabilities(data, output_file):
    """
    data = {"ID" : ("GOLD", {"LAB1" : logprob, ... })
    }    
    """
    codes = set()
    for i, (gold, probs) in data.iteritems():
        for l in probs.keys():
            codes.add(l)
    codes = sorted(codes)
    with writer(gzip.open(output_file, "w")) as ofd:
        ofd.write("\t".join(["DOC", "GOLD"] + codes) + "\n")
        for cid, (label, probs) in data.iteritems():
            ofd.write("\t".join([cid, label] + [str(probs.get(c, float("-inf"))) for c in codes]) + "\n")


def get_count(data_file):
    conc = data_file.endswith("tgz")
    ifd = CommunicationReader(data_file) if conc else reader(gzip.open(data_file))
    n = 0
    for c, _ in enumerate(ifd):
        n += 1
    logging.info("File %s contains %d Communications", data_file, n)
    return n


def extract_character_ngrams(text, n):
    stack = ["NULL" for _ in range(n)]
    ngrams = {}
    for c in text:
        stack = stack[1:]
        stack.append(c)
        ngrams[tuple(stack)] = ngrams.get(tuple(stack), 0) + 1
    return list(ngrams.iteritems())



import gzip
from io import get_count, writer

if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", dest="input")
    parser.add_argument("-o", "--output", dest="output")
    options = parser.parse_args()

    n = get_count(options.input)
    with writer(gzip.open(options.output, "w")) as ofd:
        ofd.write(str(n))



#!/usr/bin/env python

import gzip
import re
import math
import csv

pattern = re.compile(r"""
.*User\ time\ \(seconds\):\ (?P<usertime>\S+)\s+
System\ time\ \(seconds\):\ (?P<systemtime>\S+)\s+
Percent\ of\ CPU\ this\ job\ got:\ (?P<cpupercent>\S+)%\s+
Elapsed\ \(wall\ clock\)\ time\ \(h:mm:ss\ or\ m:ss\):\ (\S+)\s+
Average\ shared\ text\ size\ \(kbytes\):\ (\S+)\s+
Average\ unshared\ data\ size\ \(kbytes\):\ (\S+)\s+
Average\ stack\ size\ \(kbytes\):\ (?P<avgstack>\S+)\s+
Average\ total\ size\ \(kbytes\):\ (?P<avgtotal>\S+)\s+
Maximum\ resident\ set\ size\ \(kbytes\):\ (?P<maxresidentset>\S+)\s+
Average\ resident\ set\ size\ \(kbytes\):\ (?P<avgresidentset>\S+)\s+
Major\ \(requiring\ I/O\)\ page\ faults:\ (\S+)\s+
Minor\ \(reclaiming\ a\ frame\)\ page\ faults:\ (\S+)\s+
Voluntary\ context\ switches:\ (\S+)\s+
Involuntary\ context\ switches:\ (\S+)\s+
Swaps:\ (\S+)\s+
File\ system\ inputs:\ (\S+)\s+
File\ system\ outputs:\ (\S+)\s+
Socket\ messages\ sent:\ (\S+)\s+
Socket\ messages\ received:\ (\S+)\s+
Signals\ delivered:\ (\S+)\s+
Page\ size\ \(bytes\):\ (\S+)\s+
Exit\ status:\ (\S+).*""", re.DOTALL | re.VERBOSE)


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-o", "--output", dest="output")
    parser.add_argument("-s", "--stage", dest="stage")
    parser.add_argument(nargs="+", dest="inputs")
    options = parser.parse_args()
    
    vals = {}
    for f in options.inputs:
        task, model, size, fold, _ = re.match(r"^work/(.*?)_(.*?)_(.*?)_(.*?)(\.|_).*txt$", f).groups()
        size = int(size)
        fold = int(fold)
        with open(f) as ifd:
            key = (task, size, model, fold)
            res = {k : float(v) for k, v in pattern.search(ifd.read()).groupdict().iteritems()}
            vals[key] = (res["maxresidentset"], res["usertime"])

    with gzip.open(options.output, "w") as ofd:
        c = csv.DictWriter(ofd, fieldnames=["task", "size", "model", "fold", "%s_Memory" % (options.stage), "%s_CPU" % (options.stage)], delimiter="\t")
        c.writeheader()
        for (task, size, model, fold), (mem, cpu) in sorted(vals.iteritems()):
            c.writerow({"task" : task, "model" : model, "size" : size, "fold" : fold, "%s_Memory" % (options.stage) : mem / 1000000.0, "%s_CPU" % (options.stage) : cpu})




#!/usr/bin/env python

import gzip
import codecs
import random
from io import reader, writer

if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-i", "--input", dest="input")
    parser.add_argument("--total_file", dest="total_file")
    parser.add_argument("--training_size", dest="training_size", type=int)
    parser.add_argument("--testing_size", dest="testing_size", type=int)
    parser.add_argument("--train", dest="train")
    parser.add_argument("--test", dest="test")
    options = parser.parse_args()

    with reader(gzip.open(options.total_file)) as ifd:
        total = int(ifd.read().strip())

    indices = range(total)
    random.shuffle(indices)
    
    with writer(gzip.open(options.train, "w")) as ofd:
        ofd.write("\n".join([str(i) for i in indices[0:options.training_size]]))

    with writer(gzip.open(options.test, "w")) as ofd:
        ofd.write("\n".join([str(i) for i in indices[options.training_size:options.training_size+options.testing_size]]))





#!/usr/bin/env python

from valid.model import Compressor, Classifier
import gzip
import codecs
from itertools import chain
import pickle
import logging
import math
import numpy
from steamroller.tools.io import read_data, write_probabilities, writer, reader

if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("--input", dest="input")
    parser.add_argument("--train", dest="train")
    parser.add_argument("--dev", dest="dev")
    parser.add_argument("--test", dest="test")
    parser.add_argument("--model", dest="model")
    parser.add_argument("--output", dest="output")
    parser.add_argument("--max_ngram", dest="max_ngram", type=int, default=4)
    options = parser.parse_args()

    logging.basicConfig(level=logging.INFO)
    
    if options.train and options.output and options.input:
        compressors = {}
        counts = {}
        for cid, label, text in read_data(options.input, options.train):
            counts[label] = counts.get(label, 0) + 1
            compressors[label] = compressors.get(label, Compressor(label, options.max_ngram))
            compressors[label].add(text)

        model = Classifier()
        for l, c in compressors.iteritems():
            model.add(c, l)

        with gzip.open(options.output, "w") as ofd:
            pickle.dump((model, counts), ofd)

    # testing
    elif options.test and options.model and options.output and options.input:
        instances, gold = [], []
        for cid, label, text in read_data(options.input, options.test):
            instances.append(text)
            gold.append((cid, label))
        
        with gzip.open(options.model) as ifd:
            model, counts = pickle.load(ifd)
        total = sum(counts.values())
        prior = {k : math.log(v / float(total)) for k, v in counts.iteritems()}
        codes = model.compressors.keys()
        logging.info("Testing with %d instances, %d labels", len(instances), len(codes))
        data = {}
        for (cid, label), text in zip(gold, instances):
            probs = {k : v + prior[k] for k, v in model.classify(text).iteritems()}
            total = reduce(numpy.logaddexp, probs.values())
            data[cid] = (label, probs)
        write_probabilities(data, options.output)



import gzip
import numpy
import re
from sklearn.metrics import f1_score
import math
import csv

if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-o", "--output", dest="output")
    parser.add_argument(nargs="+", dest="inputs")
    options = parser.parse_args()
    
    scores = {}
    accuracies = {}
    for f in options.inputs:
        task, model, size, fold = re.match(r"^work/(.*)_(.*)_(.*)_(.*)_probabilities.txt.gz$", f).groups()
        size = int(size)
        fold = int(fold)
        vals = {}
        correct, incorrect = 0.0, 0.0
        truepos = 0.0
        falsepos = 0.0
        trueneg = 0.0
        falseneg = 0.0
        with gzip.open(f) as ifd:
            fields = ifd.readline().strip().split("\t")
            y_true = []
            y_pred = []
            for l in ifd:
                toks = l.strip().split("\t")
                cid, gold = toks[0:2]
                s = {k : float(v) for k, v in zip(fields[2:], toks[2:])}
                guess = sorted([(v, k) for k, v in s.iteritems()])[-1][1]
                y_true.append(gold)
                y_pred.append(guess)
                if gold == guess:
                    correct += 1.0
                else:
                    incorrect += 1.0
                        
            key = (task, size, model, fold)
            scores[key] = scores.get(key, []) + [f1_score(y_true=y_true, y_pred=y_pred, average="macro")]
            accuracies[key] = accuracies.get(key, []) + [correct / (correct + incorrect)]

    with gzip.open(options.output, "w") as ofd:
        c = csv.DictWriter(ofd, fieldnames=["task", "size", "model", "fold", "F_Score"], delimiter="\t")
        c.writeheader()
        for (task, size, model, fold), a in sorted(scores.iteritems()):
            ss = numpy.array(a)
            c.writerow({"task" : task, "model" : model, "size" : size, "fold" : fold, "F_Score" : ss.mean()})



#!/usr/bin/env python

from setuptools import setup

setup(name="SteamRoller",
      version="1.0.13",
      description="An experimental framework for text classification experiments",
      author="Tom Lippincott",
      author_email="tom@cs.jhu.edu",
      url="http://hltcoe.jhu.edu",
      maintainer="Tom Lippincott",
      maintainer_email="tom@cs.jhu.edu",
      packages=["steamroller",
                "steamroller.tools",
                "steamroller.tasks",
                "steamroller.models",
                "steamroller.metrics",
                "steamroller.plots",
                "steamroller.scons"],
      package_dir={"steamroller" : "src"},
      scripts=["scripts/steamroller"],
      install_requires=["concrete", "scikit-learn==0.18.1", "scons", "plotnine", "flask", "valid"],
      package_data={"" : ["data/SConstruct", "data/steamroller_config.json.template"]},
     )



#!/usr/bin/env python

import csv
import os.path
import os
import gzip
import re


if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-o", "--output", dest="output")
    parser.add_argument(dest="inputs", nargs="+")
    options = parser.parse_args()

    sizes = {}
    for f in options.inputs:
        task, model, size, fold = re.match(r"^work/(.*)_(.*)_(.*)_(.*)\.model.*$", f).groups()
        size = int(size)
        fold = int(fold)
        key = (task, size, model, fold)
        sizes[key] = os.lstat(f).st_size
        
    with gzip.open(options.output, "w") as ofd:
        c = csv.DictWriter(ofd, fieldnames=["task", "size", "model", "fold", "Gigabytes"], delimiter="\t")
        c.writeheader()
        for (task, size, model, fold), s in sorted(sizes.iteritems()):
            c.writerow({"task" : task, "model" : model, "size" : size, "fold" : fold, "Gigabytes" : s / 1000000000.0})



#!/usr/bin/env python

import gzip
import pickle
import codecs
import logging
import tempfile
from itertools import chain
import subprocess
import os
import shlex
import shutil
import math
import re
from steamroller.tools.io import read_data, write_probabilities, writer, reader

if __name__ == "__main__":
    
    import argparse
    
    logging.basicConfig(level=logging.INFO)
    
    parser = argparse.ArgumentParser()
    parser.add_argument("--input", dest="input")
    parser.add_argument("--train", dest="train")
    parser.add_argument("--test", dest="test")
    parser.add_argument("--model", dest="model")
    parser.add_argument("--output", dest="output")
    
    parser.add_argument("--min_char_ngram", dest="min_char_ngram", type=int, default=0)
    parser.add_argument("--max_char_ngram", dest="max_char_ngram", type=int, default=4)
    
    parser.add_argument("--word_vector_size", dest="word_vector_size", type=int, default=100)
    parser.add_argument("--word_context_size", dest="word_context_size", type=int, default=5)
    parser.add_argument("--epochs", dest="epochs", type=int, default=100)
    parser.add_argument("--negatives_sampled", dest="negatives_sampled", type=int, default=5)
    parser.add_argument("--min_word_count", dest="min_word_count", type=int, default=0)
    parser.add_argument("--min_label_count", dest="min_label_count", type=int, default=0)
    parser.add_argument("--max_word_ngram", dest="max_word_ngram", type=int, default=1)
    parser.add_argument("--bucket_count", dest="bucket_count", type=int, default=2000000)
    parser.add_argument("--threads", dest="threads", type=int, default=1)
    parser.add_argument("--verbosity", dest="verbosity", type=int, default=1)
    
    parser.add_argument("--learning_rate", dest="learning_rate", type=float, default=0.1)
    parser.add_argument("--learning_rate_update_rate", dest="learning_rate_update_rate", type=int, default=100)
    parser.add_argument("--sampling_threshold", dest="sampling_threshold", type=float, default=0.0001)
    
    parser.add_argument("--label_prefix", dest="label_prefix", default="__label__")
    parser.add_argument("--loss_function", dest="loss_function", choices=["ns", "hs", "softmax"], default="ns")
    options = parser.parse_args()

    args = {k : v for k, v in options._get_kwargs()}
    train_command = "fasttext supervised -input %(input_file)s -output %(output_file)s -lr %(learning_rate)f -lrUpdateRate %(learning_rate_update_rate)d -dim %(word_vector_size)d -ws %(word_context_size)d -epoch %(epochs)d -minCount %(min_word_count)d -minCountLabel %(min_label_count)d -neg %(negatives_sampled)d -wordNgrams %(max_word_ngram)d -loss %(loss_function)s -bucket %(bucket_count)d -minn %(min_char_ngram)d -maxn %(max_char_ngram)d -thread %(threads)d -t %(sampling_threshold)f -label %(label_prefix)s -verbose %(verbosity)d"
    apply_command = "fasttext predict-prob %(model)s %(test_file)s %(num_labels)d"

    # training
    if options.train and options.output and options.input:
        data = read_data(options.input, options.train)
        instances = [x[2] for x in data]
        _, model_fname = tempfile.mkstemp(suffix=".bin")
        base_fname = os.path.splitext(model_fname)[0]
        vec_fname = "%s.vec" % (base_fname)
        _, input_fname = tempfile.mkstemp()
        label_lookup = {}
        for _, l, _ in data:
            label_lookup[l] = label_lookup.get(l, len(label_lookup))
        with writer(open(input_fname, "w")) as ofd:
            for _, label, text in data:
                ofd.write("__label__%s %s\n" % (label, text))
        logging.info("Training with %d instances, %d labels", len(data), len(label_lookup))
        args["input_file"] = input_fname
        args["output_file"] = base_fname
        toks = shlex.split(train_command % args)
        p = subprocess.Popen(toks)
        p.communicate()
        os.remove(input_fname)
        shutil.copyfile(model_fname, options.output)
        os.remove(model_fname)

    # testing
    elif options.test and options.model and options.output and options.input:
        instances, gold = [], []
        labels = set()
        _, test_fname = tempfile.mkstemp()
        with writer(open(test_fname, "w")) as ofd:
            for cid, label, text in read_data(options.input, options.test):
                instances.append(text)
                gold.append((cid, label))
                ofd.write(text + "\n")
                labels.add(label)
        
        logging.info("Testing with %d instances", len(instances))
        args["num_labels"] = 100
        args["test_file"] = test_fname
        p = subprocess.Popen(shlex.split(apply_command % args), stdout=subprocess.PIPE)
        out, err = p.communicate()
        os.remove(test_fname)
        
        data = {}
        for line, (cid, label) in zip(out.strip().split("\n"), gold):
            probs = {k : math.log(float(v)) for k, v in [m.groups() for m in re.finditer(r"__label__(\S+) (\S+)", line)]}
            for k in probs.keys():
                labels.add(k)
            data[cid] = (label, probs)
        write_probabilities(data, options.output)
    else:
        print "ERROR: you must specify --input and --output, and either --train or --test and --model!"



#!/usr/bin/env python



#
# Data-loading functions (can probably skip this)
#


def instances_and_lookups(input_file, index_file, sym_lookup={"unk" : 0}, label_lookup={"unk" : 0}):
    """
    Read communications and create integer encodings for them, along with lookups to recover the
    strings.  "unk" is mapped to 0 for both symbols and labels, to handle OOV at test time.  If
    symbol or label lookups are passed to the function, does *not* update the lookups and encodes
    unseen items as "unk".

    In other words, when reading training data, don't pass lookups.  When reading test data, pass in
    the lookups from the training data.
    """
    assert(sym_lookup["unk"] == 0 and label_lookup["unk"] == 0)
    update_sym = len(sym_lookup) == 1
    update_label = len(label_lookup) == 1
    cid_lookup = {}
    instances, labels = [], []
    unk_sym_occs, unk_sym_types = 0, set()
    unk_label_occs, unk_label_types = 0, set()
    for cid, label, text in read_data(options.input, options.train if options.train else options.test):
        if update_label:
            label_lookup[label] = label_lookup.get(label, len(label_lookup))
        cid_lookup[cid] = label_lookup.get(cid, len(cid_lookup))
        syms = []
        for c in text:
            if update_sym:
                sym_lookup[c] = sym_lookup.get(c, len(sym_lookup))
            syms.append(sym_lookup.get(c, 0))
            if syms[-1] == 0:
                unk_sym_types.add(c)
                unk_sym_occs += 1
        instances.append((label_lookup.get(label, 0), syms, cid_lookup[cid]))
        if instances[-1][0] == 0:
            unk_sym_types.add(label)
            unk_label_occs += 1            
    logging.info("Loaded %d instances, %d labels", len(instances), len(label_lookup))
    logging.info("%d/%d unknown symbol occurrences/types, %d/%d unknown label occurences/types",
                 unk_sym_occs,
                 len(unk_sym_types),
                 unk_label_occs,
                 len(unk_label_types),                 
    )
    return instances, cid_lookup, sym_lookup, label_lookup


def example_generator(examples):
    """
    Simply yields examples per TF idiom.
    """
    for example in examples:
        yield {'text': example[1], 'label': example[0], "cid" : example[2]}


def serialize_model(sess, temppath, sym_lookup, label_lookup, cid_lookup, output_file):
    with gzip.open(os.path.join(temppath, "lookups.pkl.gz"), "w") as ofd:
        pickle.dump((sym_lookup, label_lookup, cid_lookup), ofd)
    saver = tf.train.Saver()
    saver.save(sess, os.path.join(temppath, "model"))
    #with open(os.path.join(checkpoint_path, "checkpoint")) as ifd:
    #    text = ifd.read()
    #    text = re.sub(checkpoint_path, "TMPPATH", text)
    #with open(os.path.join(checkpoint_path, "checkpoint"), "w") as ofd:
    #    ofd.write(text)
    with tarfile.open(output_file, "w:gz") as ofd:
        for name in glob(os.path.join(temppath, "*")):
            ofd.add(name, arcname=os.path.basename(name))


def deserialize_model(model_file, output_path):
    with tarfile.open(model_file) as ifd:
        ifd.extractall(path=output_path)
    #with open(os.path.join(output_path, "checkpoint")) as ifd:
    #    text = ifd.read()
    #    text = re.sub("TMPPATH", temppath, text)
    #with open(os.path.join(output_path, "checkpoint"), "w") as ofd:
    #    ofd.write(text)
    with gzip.open(os.path.join(output_path, "lookups.pkl.gz")) as ifd:
        sym_lookup, label_lookup, _ = pickle.load(ifd)
    return sym_lookup, label_lookup


#
# Callbacks for inference mode (irrelevant to training)
#


def key_callback(id_to_label, id_to_cid, x):
    """
    When TF iterates over examples, map the labels and IDs from integers back to strings.
    """
    return (id_to_label[x["label"]], id_to_cid[x["cid"]])


def result_callback(id_to_label, output_file, res):
    """
    When TF inference completes, assemble the results to write to disk.
    """
    data = {}
    order = [id_to_label[i] for i in range(len(id_to_label))]
    correct, total = 0, 0
    for (gold, cid), (probs,) in res:
        total += 1
        if gold == order[probs.flatten().argmax()]:
            correct += 1
        data[cid] = (gold, {k : v for k, v in zip(order, probs.flatten())})
    logging.info("Accuracy: %.3f", float(correct) / total)
    write_probabilities(data, options.output)

        
#
# Modified Seal code
#
# General goals: remove global variables as much as possible, and annotate
#                code with what I believe it does.  Remove and consolidate
#                things to a minimal functional flat RNN.
#

class SequenceModel(object):

    def __init__(self, num_letters, num_labels, char_rnn_layer, char_state_vector_length, char_embed_vector_length):
        logging.info("Creating model with %d letters and %d labels", num_letters, num_labels)
        #self._keep_probability = tf.Variable(1.0, trainable=False)
        self._keep_probability = tf.placeholder_with_default(1.0, [], name='keep_probability')
        self._char_cells = []
        for l in range(char_rnn_layer):
            char_cell = tf.contrib.rnn.GRUCell(num_units=char_state_vector_length)
            char_cell = tf.contrib.rnn.DropoutWrapper(char_cell,
                                                      input_keep_prob=self._keep_probability,
                                                      output_keep_prob=self._keep_probability)
            char_cell = td.ScopedLayer(char_cell, 'char_cell_{}'.format(l))
            self._char_cells.append(char_cell)

        # Character-level stacked RNN
        self._char_embed = (td.Map(td.Scalar(tf.int64) >>
                             td.Function(td.Embedding(num_letters,
                                                      char_embed_vector_length,
                                                      name='char_embed'))))
        self._char_rnn = self._char_embed >> td.RNN(self._char_cells[0])
        for l in range(1, char_rnn_layer):
            self._char_rnn >>= td.GetItem(0) >> td.RNN(self._char_cells[l])

        self._context = self._char_rnn >> td.GetItem(1)    

        self._output_layer = self._context >> td.FC(num_labels,
                                                    activation=None,
                                                    input_keep_prob=self._keep_probability,
                                                    name='output_layer')

        self._label = td.Scalar(tf.int64)
        self._root_block = td.Record([('text', self._output_layer),
                                      ('label', self._label)])

        self._compiler = td.Compiler.create(self._root_block)
        (self._logits, self._labels) = self._compiler.output_tensors

        self._loss = tf.nn.sparse_softmax_cross_entropy_with_logits(
            logits=self._logits,
            labels=self._labels
        )
        self._global_step = tf.contrib.framework.get_or_create_global_step()
        lr = tf.train.exponential_decay(.001, self._global_step,
                                    100000, 0.96, staircase=True)

        self._optimizer = build_optimizer_from_params("adam", learning_rate=.001)
        tvars = tf.trainable_variables()

        self._predictions = tf.argmax(self._logits, 1)
        self._accuracy = tf.reduce_mean(
            tf.cast(tf.equal(self._predictions, self._labels), dtype=tf.float32)
        )

        self._grads_and_vars = self._optimizer.compute_gradients(self._loss, tvars)

        tf.clip_by_global_norm([x[0] for x in self._grads_and_vars], 10.0)
        self._train_op = self._optimizer.minimize(self._loss, global_step=self._global_step)

    @property
    def loss(self):
        return self._loss

    @property
    def accuracy(self):
        return self._accuracy

    @property
    def compiler(self):
        return self._compiler

    @property
    def train_op(self):
        return self._train_op

    @property
    def global_step(self):
        return self._global_step

    @property
    def predictions(self):
        return self._predictions

    @property
    def labels(self):
        return self._labels

    @property
    def log_probs(self):
        return self._logits
    
    @property
    def outputs(self):
        return [self._logits]
    
    def build_feed_dict(self, expressions):
        return self._compiler.build_feed_dict(expressions)


#
# Entrypoint for script
#


if __name__ == "__main__":
    import re
    import gzip
    import pickle
    import codecs
    import logging
    from functools import partial
    from itertools import chain
    import random
    import tempfile
    import shutil
    import os.path
    import tarfile
    from glob import glob
    import tensorflow as tf
    from tensorflow.python.ops import math_ops, array_ops
    from tensorflow.python.ops import variable_scope as vs
    import tensorflow.contrib.seq2seq as seq2seq
    import tensorflow_fold as td
    import tensorflow_fold.blocks.blocks
    import tensorflow_fold.blocks.result_types as tdt
    from tensorflow_fold.blocks.plan import build_optimizer_from_params
    from tqdm import tqdm
    import numpy
    from steamroller.tools.io import read_data, write_probabilities, writer, reader, extract_character_ngrams

    import argparse

    logging.basicConfig(level=logging.INFO)
    
    parser = argparse.ArgumentParser()
    parser.add_argument("--input", dest="input")
    parser.add_argument("--train", dest="train")
    parser.add_argument("--test", dest="test")
    parser.add_argument("--model", dest="model")
    parser.add_argument("--output", dest="output")
    
    parser.add_argument("--dev", dest="dev", type=float, default=.1)
    parser.add_argument("--char_embed_vector_length", dest="char_embed_vector_length", default=512, type=int)
    parser.add_argument("--char_state_vector_length", dest="char_state_vector_length", default=512, type=int)
    parser.add_argument("--char_rnn_layer", dest="char_rnn_layer", default=2, type=int)
    parser.add_argument("--keep_probability", dest="keep_probability", default=.5, type=float)
    parser.add_argument("--max_grad_norm", dest="max_grad_norm", default=10.0, type=float)
    parser.add_argument("--learning_rate", dest="learning_rate", default=.001, type=float)

    parser.add_argument("--epochs", dest="epochs", default=100, type=int)
    parser.add_argument("--batch_size", dest="batch_size", default=32, type=int)
    parser.add_argument("--batches_per_epoch", dest="batches_per_epoch", default=100, type=int)
    options = parser.parse_args()
    
    temppath = tempfile.mkdtemp(prefix="tlippincott-tf")

    try:

        # train model
        if options.train and options.output and options.input:
            instances, cid_lookup, sym_lookup, label_lookup = instances_and_lookups(options.input,
                                                                                    options.train)
            random.shuffle(instances)
            train = instances[0 : int(len(instances) * (1.0 - options.dev))]
            dev = instances[int(len(instances) * (1.0 - options.dev)):]
            best_score = None
            with tf.Graph().as_default():
                model = SequenceModel(len(sym_lookup),
                                      len(label_lookup),
                                      options.char_rnn_layer,
                                      options.char_state_vector_length,
                                      options.char_embed_vector_length,
                )
                #init = tf.assign(model._keep_probability, options.keep_probability)
                init = tf.global_variables_initializer()
                #supervisor = tf.train.Supervisor(logdir=temppath)
                #with supervisor.managed_session() as sess:
                with tf.Session() as sess:
                    sess.run(init)
                    since_last_write = 0
                    #sess.run(init) #tf.assign(model._keep_probability, options.keep_probability))
                    for e in xrange(options.epochs):
                        logging.info("Epoch %d", e + 1)
                        random.shuffle(train)
                        accs = []
                        for i, batch_feed in enumerate(model.compiler.build_loom_input_batched(example_generator(train), options.batch_size)):
                            if i > options.batches_per_epoch:
                                break
                            else:
                                _, step, loss_v, acc_v = sess.run(
                                    [model.train_op, model.global_step, model.loss, model.accuracy],
                                    feed_dict={model.compiler.loom_input_tensor : batch_feed, model._keep_probability : options.keep_probability})
                                accs.append(acc_v)
                        logging.info("Train accuracy over epoch #%d was %f", e + 1, sum(accs) / len(accs))
                        accs = []
                        for batch_feed in model.compiler.build_loom_input_batched(example_generator(dev), options.batch_size):
                            acc, preds, prob = sess.run([model.accuracy, model.predictions, model.log_probs],
                                                        feed_dict={model.compiler.loom_input_tensor : batch_feed, model._keep_probability : 1.0})
                            accs.append(acc)
                            #probs.append(prob)
                        acc = sum(accs) / len(accs)
                        logging.info("Dev accuracy over epoch #%d was %f", e + 1, acc)
                        since_last_write += 1
                        if since_last_write > 5:
                            logging.info("Stopping!")
                            break
                        if best_score == None or acc > best_score:
                            logging.info("New best score: %f", acc)
                            best_score = acc
                            since_last_write = 0
                            serialize_model(sess, temppath, sym_lookup, label_lookup, cid_lookup, options.output)
            
        # perform inference with existing model
        elif options.test and options.model and options.output and options.input:
            train_sym_lookup, train_label_lookup = deserialize_model(options.model, temppath)
            instances, cid_lookup, sym_lookup, label_lookup = instances_and_lookups(options.input,
                                                                                    options.train,
                                                                                    sym_lookup=train_sym_lookup,
                                                                                    label_lookup=train_label_lookup)
            
            random.shuffle(instances)
            id_to_label = {v : k for k, v in label_lookup.iteritems()}
            id_to_cid = {v : k for k, v in cid_lookup.iteritems()}
            test = example_generator(instances)
            
            with tf.Graph().as_default():
                model = SequenceModel(len(sym_lookup),
                                      len(label_lookup),
                                      options.char_rnn_layer,
                                      options.char_state_vector_length,
                                      options.char_embed_vector_length,
                )
                #init = tf.global_variables_initializer()
                #init = tf.assign(model._keep_probability, options.keep_probability)
                #supervisor = tf.train.Supervisor(logdir=temppath)
                data = {}
                #with supervisor.managed_session() as sess:
                with tf.Session() as sess:
                    s = tf.train.Saver()
                    s.restore(sess, os.path.join(temppath, "model"))
                    #sess.run(init) #tf.assign(model._keep_probability, 1.0))
                    accs = []
                    probs = []
                    for batch_feed in model.compiler.build_loom_input_batched(test, options.batch_size):
                        
                        acc, preds, prob = sess.run([model.accuracy, model.predictions, model.log_probs],
                                                     feed_dict={model.compiler.loom_input_tensor : batch_feed, model._keep_probability : 1.0})
                        accs.append(acc)
                        probs.append(prob)
                probs = numpy.concatenate(probs, axis=0)
                logging.info("Accuracy: %f", sum(accs) / len(accs))
                data = {}
                order = [id_to_label[i] for i in range(len(id_to_label))]
                for dist, (lid, _, cid) in zip(probs, instances):
                    cid = id_to_cid.get(cid, str(cid))
                    g = id_to_label[lid]
                    data[cid] = (g, {k : v for k, v in zip(order, dist.flatten())})

                write_probabilities(data, options.output)
        else:
            logging.error("You must specify the input and output files, and either a training file, or testing and model files!")
    finally:
        shutil.rmtree(temppath)



#!/usr/bin/env python

import gzip
from sklearn import naive_bayes, linear_model, svm, metrics, ensemble, dummy
from sklearn.feature_extraction import DictVectorizer
import pickle
import codecs
import logging
from itertools import chain
from steamroller.tools.io import read_data, write_probabilities, writer, reader, extract_character_ngrams

models = {
    "naive_bayes" : (naive_bayes.MultinomialNB, {}, []),
    "svm" : (svm.LinearSVC, {}, ["C"]),
    "logistic_regression" : (linear_model.LogisticRegression, {"class_weight" : "balanced"}, ["C"]),
    "random_forest" : (ensemble.RandomForestClassifier, {"class_weight" : "balanced"}, []),
    "prior" : (dummy.DummyClassifier, {"strategy" : "prior"}, []),
}


if __name__ == "__main__":

    import argparse

    logging.basicConfig(level=logging.INFO)
    
    parser = argparse.ArgumentParser()
    parser.add_argument("--input", dest="input")
    parser.add_argument("--type", dest="type", choices=models.keys())
    parser.add_argument("--train", dest="train")
    parser.add_argument("--test", dest="test")
    parser.add_argument("--model", dest="model")
    parser.add_argument("--output", dest="output")
    parser.add_argument("--max_ngram", dest="max_ngram", type=int, default=4)    
    options = parser.parse_args()
    
    # training
    if options.train and options.output and options.input:
        instances, labels = [], []        
        for cid, label, text in read_data(options.input, options.train):
            instances.append(dict(sum([extract_character_ngrams(text, n) for n in range(1, options.max_ngram + 1)], [])))
            labels.append(label)
        dv = DictVectorizer(sparse=True)
        X = dv.fit_transform(instances)
        label_lookup = {}
        classifier_class, args, hypers = models[options.type]
        classifier = classifier_class(**args)
        for l in labels:
            label_lookup[l] = label_lookup.get(l, len(label_lookup))
        logging.info("Training with %d instances, %d features, %d labels", len(instances), len(dv.get_feature_names()), len(label_lookup))
        classifier.fit(X, [label_lookup[l] for l in labels])
        with gzip.open(options.output, "w") as ofd:
            pickle.dump((classifier, dv, label_lookup), ofd)            
    # testing
    elif options.test and options.model and options.output and options.input:
        with gzip.open(options.model) as ifd:
            classifier, dv, label_lookup = pickle.load(ifd)
        instances, gold = [], []
        for cid, label, text in read_data(options.input, options.test):
            instances.append(dict(sum([extract_character_ngrams(text, n) for n in range(1, options.max_ngram + 1)], [])))
            gold.append((cid, label))
        logging.info("Testing with %d instances, %d labels", len(instances), len(label_lookup))
        X = dv.transform(instances)
        inv_label_lookup = {v : k for k, v in label_lookup.iteritems()}
        data = {}
        order = [inv_label_lookup[i] for i in range(len(inv_label_lookup))]
        if hasattr(classifier, "predict_log_proba"):
            for probs, (cid, g) in zip(classifier.predict_log_proba(X), gold):
                data[cid] = (g, {k : v for k, v in zip(order, probs.flatten())})
        else:
            for pred, (cid, g) in zip(classifier.predict(X), gold):
                probs = [0.0 if i == pred else float("-inf") for i in range(len(order))]
                data[cid] = (g, {k : v for k, v in zip(order, probs)})
        write_probabilities(data, options.output)
    else:
        print "ERROR: you must specify --input and --output, and either --train or --test and --model!"



import pandas
import numpy
import gzip
import csv
import matplotlib as mpl
mpl.use('Agg')
from plotnine import ggplot, aes, geom_boxplot, coord_flip, theme, geom_violin, geom_point, scale_x_continuous, ylab, xlab, stat_smooth, geom_line, ggtitle, guide_legend, theme_minimal, theme_void, theme_update, element_text, labs
from plotnine.themes.themeable import legend_title
from plotnine.guides import guide_legend, guides
import re

def maybe_float(s):
    try:
        return float(s)
    except:
        return None

if __name__ == "__main__":

    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument("-t", "--title", dest="title", default="Box Plot")
    parser.add_argument("-o", "--output", dest="output")
    parser.add_argument("-i", "--input", dest="input")
    parser.add_argument("--x", dest="x")
    parser.add_argument("--y", dest="y")
    parser.add_argument("--color", dest="color")
    parser.add_argument("--xlabel", dest="xlabel")
    parser.add_argument("--ylabel", dest="ylabel")
    parser.add_argument("--color_label", dest="color_label")
    options = parser.parse_args()

    items = {}
    data = {}
    all_models = set()
    all_tasks = set()
    data = {}
    with gzip.open(options.input) as ifd:
        for row in csv.DictReader(ifd, delimiter="\t"):
            for k, v in row.iteritems():
                data[k] = data.get(k, [])
                data[k].append(v)
    
    for k in data.keys():
        floats = [maybe_float(x) for x in data[k]]
        if all([re.match(r"^\d+$", x) for x in data[k]]):
            data[k] = [int(x) for x in data[k]]
        elif all(floats):
            data[k] = floats

    df = pandas.DataFrame(data)
    x = (ggplot(df, aes("factor(%s)" % (options.x), options.y, color="factor(%s)" % (options.color)))) + \
        theme(legend_title=element_text("")) + \
        ggtitle(options.title.strip("'")) + \
        ylab(options.ylabel.strip("'")) + \
        xlab(options.xlabel.strip("'")) + \
        labs(color=options.color_label.strip("'")) + \
        geom_boxplot()
    x.save(options.output)
